from typing import Optional

import envium
from envium import env_var

__all__ = ["Environ", "environ", "TestType"]


class TestType:
    UNIT = "unit"
    E2E = "e2e"


class Environ(envium.Environ):
    test_type: Optional[str] = env_var()
    stickybeak: bool = env_var(default=False)
    colors: bool = env_var(default=True)
    stage: Optional[str] = env_var()


environ = Environ(name="envo", load=True)



import re
import sys
import traceback
from dataclasses import dataclass, field
from enum import Enum
from pathlib import Path
from typing import Any, Dict, List, Optional

from pygments.styles import get_style_by_name
from rhei import Stopwatch
from xonsh.pyghooks import XonshConsoleLexer

from envo.environ import environ
from envo.vendored.colorizer import Colorizer, AnsiParser
from envo import console


class Level(int, Enum):
    DEBUG = 0
    INFO = 1
    WARNING = 2
    ERROR = 3


@dataclass
class Msg:
    level: Level
    body: str
    time: float  # s
    descriptor: Optional[str] = None
    metadata: Dict[str, Any] = field(default_factory=dict)

    level_to_format = {
        Level.DEBUG: lambda m: f"<blue>{m}</blue>",
        Level.INFO: lambda m: f"<bold>{m}</bold>",
        Level.WARNING: lambda m: f"<bold><yellow>{m}</yellow></bold>",
        Level.ERROR: lambda m: f"<bold><red>{m}</red></bold>"
    }

    def __post_init__(self) -> None:
        self.body = str(self.body).lstrip()

    def print(self, add_color: bool = True, add_time: bool = False, add_level: bool = False, highlight_body: bool = False) -> None:
        msg = self.render(add_color=add_color, add_time=add_time, add_level=add_level, highlight_body=highlight_body)
        if self.level >= Level.WARNING:
            sys.stderr.write(msg + "\n")
            sys.stderr.flush()
        else:
            print(msg)

    def _fix_formatting(self, text: str) -> str:
        parser = AnsiParser()

        for match in parser._regex_tag.finditer(text):
            markup, tag = match.group(0), match.group(1)

            if tag not in {"lvl", "level"}:
                ansi = parser._get_ansicode(tag)
                if ansi is None:
                    text = text.replace(markup, markup.replace("<", "< ").replace(">", " >"))

        return text

    def render(self, add_color: bool = True, add_time: bool = False, add_level: bool = False, highlight_body: bool = False) -> str:
        if not environ.colors:
            add_color = False

        def colored(clr: str) -> str:
            if add_color:
                return clr
            else:
                return ""

        from pygments import highlight
        from pygments.formatters.terminal import TerminalFormatter

        parts = []

        if add_level:
            parts.append(f"[{self.time:.4f}]")

        if add_time:
            parts.append(f"[{self.level.name:<5}]")

        if add_level or add_time:
            parts.append(" ")

        if self.descriptor:
            descriptor = f"{colored('<green>')}({str(self.descriptor)}) {colored('</green>')}"
            parts.append(f"{descriptor}")

        if add_color:
            if highlight_body:
                lines = []
                for l in self.body.splitlines():
                    spaces = len(l) - len(l.lstrip())
                    l = l.lstrip()
                    l = self._fix_formatting(l)
                    l = highlight(
                        l,
                        XonshConsoleLexer(),
                        TerminalFormatter(style=get_style_by_name("emacs")),
                    )
                    lines.append(" " * spaces + l)
                parts.append("".join(lines))
            else:
                part = self.level_to_format[self.level](self.body)
                parts.append(part)
        else:
            parts.append(self.body)

        if self.metadata:
            parts.append(f"; {self.metadata}")

        msg = "".join(parts)
        msg = Colorizer.ansify(msg)
        msg = msg.strip()
        return msg

    def __repr__(self) -> str:
        return self.render(add_color=False)


@dataclass
class MsgFilter:
    level: Optional[Level] = None
    body_re: Optional[str] = None
    time_later: Optional[float] = None
    time_before: Optional[float] = None
    metadata_re: Optional[Dict[str, Any]] = None

    # matchers
    def matches_level(self, msg: Msg) -> bool:
        return self.level is None or msg.level == self.level

    def matches_body(self, msg: Msg) -> bool:
        return self.body_re is None or bool(re.match(self.body_re, msg.body, re.DOTALL))

    def matches_time_later(self, msg: Msg) -> bool:
        return self.body_re is None or msg.time >= self.time_later

    def matches_time_before(self, msg: Msg) -> bool:
        return self.body_re is None or msg.time < self.time_before

    def matches_metadata(self, msg: Msg) -> bool:
        if not self.metadata_re:
            return True

        for k, v in self.metadata_re.items():
            msg_value = msg.metadata.get(k)
            if msg_value is None:
                return False

            if not re.match(v, msg_value, re.DOTALL):
                return False

        return True

    def matches_all(self, msg: Msg) -> bool:
        return (
            self.matches_level(msg)
            and self.matches_body(msg)
            and self.matches_time_later(msg)
            and self.matches_time_before(msg)
            and self.matches_metadata(msg)
        )


class Messages(list):
    content = List[Msg]

    def print(self) -> None:
        ret = []

        for m in self:
            m.print()

        ret = "\n".join(ret)

        print(ret)


@dataclass
class Logger:
    name: str
    parent: Optional["Logger"] = None
    descriptor: Optional[str] = None
    log_exception_repr: bool = False

    colors: bool = True

    messages: Messages = field(init=False, default_factory=Messages)
    level: Level = field(init=False, default=Level.INFO)
    sw: Stopwatch = field(init=False, default_factory=Stopwatch)

    def __post_init__(self) -> None:
        self.sw.start()

    def create_child(self, name: str, descriptor: str) -> "Logger":
        logger = Logger(parent=self, name=name, descriptor=descriptor)
        logger.sw = self.sw
        return logger

    def clean(self) -> None:
        self.messages = Messages()

    def set_level(self, level: Level) -> None:
        self.level = level

    def _log(self, msg: Msg) -> None:
        self.messages.append(msg)

    def log(self, message: str, level: Level, metadata: Optional[Dict[str, Any]] = None, print_msg: bool = True, highlight_body: bool = False) -> None:
        msg = Msg(
            level,
            message,
            self.sw.value,
            metadata=metadata or {},
            descriptor=self.descriptor,
        )

        if level >= self.level and print_msg:
            msg.print(highlight_body=highlight_body)

        self._log(msg)

        if self.parent:
            self.parent._log(msg)

    def debug(self, message: str, metadata: Optional[Dict[str, Any]] = None) -> None:
        self.log(message, Level.DEBUG, metadata)

    def info(self, message: str, metadata: Optional[Dict[str, Any]] = None) -> None:
        self.log(message, Level.INFO, metadata)

    def warning(self, message: str, metadata: Optional[Dict[str, Any]] = None) -> None:
        self.log(message, Level.WARNING, metadata)

    def error(self, message: str, metadata: Optional[Dict[str, Any]] = None) -> None:
        self.log(message, Level.ERROR, metadata)

    def traceback(self) -> None:
        from envo.exceptions import user_code
        with user_code():
            msg = ''.join(traceback.format_exc()).strip()

        self.log(msg, level=Level.ERROR, highlight_body=True)

    def get_msgs(self, filter: MsgFilter) -> List[Msg]:
        filtered: List[Msg] = []
        for m in self.messages:
            if filter.matches_all(m):
                filtered.append(m)

        return filtered

    def print_all(self, color: bool = True) -> None:
        for m in self.messages:
            m.print(add_color=color, add_time=True, add_level=True)

    def tail(self, messages_n: int) -> None:
        for m in self.messages[-messages_n:]:
            m.print()

    def save(self, file: Path) -> None:
        pass


logger = Logger(name="root")



import importlib.machinery
import importlib.util
import sys
from contextlib import contextmanager
from pathlib import Path
from types import FrameType
from typing import Any, Callable, Dict, Generator, List, Optional, Tuple, Union


__all__ = [
    "import_from_file",
    "add_source_roots",
    "get_repo_root",
    "dir_name_to_class_name",
    "get_src_root",
    "render_py_file",
    "render_file",
    "colored",
]

from envo.logs import logger


class Callback:
    def __init__(self, func: Optional[Callable[..., Any]] = None) -> None:
        self.func = func

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        from envo.exceptions import user_code

        if not self.func:
            return
        with user_code():
            return self.func(*args, **kwargs)

    def __bool__(self) -> bool:
        return self.func is not None


def dir_name_to_class_name(dir_name: str) -> str:
    class_name = dir_name.replace("_", " ")
    class_name = class_name.replace("-", " ")
    class_name = class_name.replace(".", " ")
    s: str
    class_name = "".join([s.strip().capitalize() for s in class_name.split()])

    return class_name


def dir_name_to_pkg_name(dir_name: str) -> str:
    pkg_name = dir_name.replace("_", " ")
    class_name = pkg_name.replace("-", " ")
    class_name = class_name.replace(".", " ")
    s: str
    class_name = "_".join([s.strip() for s in class_name.split()])

    return class_name


def is_valid_module_name(module: str) -> bool:
    from keyword import iskeyword

    return module.isidentifier() and not iskeyword(module)


def render_file(template_path: Path, output: Path, context: Dict[str, Any]) -> None:
    content = template_path.read_text()
    for n, v in context.items():
        content = content.replace(f"{{{{ {n} }}}}", v)

    output.write_text(content, encoding="utf-8")


def render_py_file(template_path: Path, output: Path, context: Dict[str, Any]) -> None:
    render_file(template_path, output, context)


def path_to_module_name(path: Path, package_root: Path) -> str:
    rel_path = path.resolve().absolute().relative_to(package_root.resolve())
    ret = str(rel_path).replace(".py", "").replace("/", ".").replace("\\", ".")
    ret = ret.replace(".__init__", "")
    return ret


def import_from_file(path: Union[Path, str]) -> Any:
    path = Path(path)
    loader = importlib.machinery.SourceFileLoader(str(path), str(path))
    spec = importlib.util.spec_from_loader(loader.name, loader)
    module = importlib.util.module_from_spec(spec)
    loader.exec_module(module)

    return module


def import_env_from_file(path: Union[Path, str]) -> Any:
    ret = import_from_file(path)

    return ret


def get_module_from_full_name(full_name: str) -> Optional[str]:
    parts = full_name.split(".")

    while True:
        module_name = ".".join(parts)
        if module_name in sys.modules:
            return module_name
        parts.pop(0)
        if not parts:
            return None


PLATFORM_WINDOWS = "windows"
PLATFORM_LINUX = "linux"
PLATFORM_BSD = "bsd"
PLATFORM_DARWIN = "darwin"
PLATFORM_UNKNOWN = "unknown"


def get_platform_name():
    if sys.platform.startswith("win"):
        return PLATFORM_WINDOWS
    elif sys.platform.startswith("darwin"):
        return PLATFORM_DARWIN
    elif sys.platform.startswith("linux"):
        return PLATFORM_LINUX
    elif sys.platform.startswith(("dragonfly", "freebsd", "netbsd", "openbsd", "bsd")):
        return PLATFORM_BSD
    else:
        return PLATFORM_UNKNOWN


__platform__ = get_platform_name()


def is_linux():
    return __platform__ == PLATFORM_LINUX


def is_bsd():
    return __platform__ == PLATFORM_BSD


def is_darwin():
    return __platform__ == PLATFORM_DARWIN


def is_windows():
    return __platform__ == PLATFORM_WINDOWS


def add_source_roots(paths: List[Union[Path, str]]) -> None:
    logger.debug(f"Adding source roots {paths}")
    for p in paths:
        if not str(p).strip():
            continue

        if str(p) in sys.path:
            sys.path.remove(str(p))

        sys.path.insert(0, str(p))


def get_src_root(name: str, file: Optional[str] = None) -> Path:
    root = Path(file).parent if file else "."
    path = Path(root).absolute()

    while not list(path.glob(f"*{name}")):
        if path == path.parent:
            raise RuntimeError(f'Cannot find src root (missing "{name}" file?)')
        path = path.parent

    return path


def get_repo_root() -> Path:
    return get_src_root(".git")


def iterate_frames(frame: FrameType) -> Generator[FrameType, None, None]:
    current_frame: Optional[FrameType] = frame
    while current_frame:
        yield current_frame
        current_frame = current_frame.f_back


def colored(inp: str, color: Tuple[int, int, int]) -> str:
    ret = f"\033[38;2;{color[0]};{color[1]};{color[2]}m{inp}\x1b[0m"
    return ret


@contextmanager
def mocked(obj: Any, attr_name: str, target: Any) -> Generator[None, None, None]:
    orig = getattr(obj, attr_name)
    setattr(obj, attr_name, target)
    yield
    setattr(obj, attr_name, orig)



import builtins
import inspect
import os
import re
import sys
import traceback
from abc import ABC, abstractmethod
from collections import OrderedDict
from contextlib import contextmanager
from copy import copy, deepcopy
from dataclasses import dataclass, field, is_dataclass
from functools import wraps
from pathlib import Path
from threading import Lock, Thread
from time import sleep
from types import FrameType, MethodType, ModuleType
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    ClassVar,
    Dict,
    Generic,
    List,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
)

import envium
from envium import computed_env_var, env_var
from rhei import Stopwatch

from envo import logger, utils
from envo.environ import environ
from envo.exceptions import user_code
from envo.logs import Logger
from envo.status import Status
from envo.utils import (
    Callback,
    import_env_from_file,
    import_from_file,
)

__all__ = [
    "Env",
    "command",
    "shell_context",
    "precmd",
    "postcmd",
    "onstdout",
    "onstderr",
    "oncreate",
    "onload",
    "onunload",
    "ondestroy",
    "boot_code",
    "Source",
]

T = TypeVar("T")

from envo.exceptions import UserError
from envo.shell import Shell

if TYPE_CHECKING:
    from envo import Plugin
    from envo.scripts import Status


class MagicFunctionData:
    type: str
    namespace: str
    expected_fun_args: List[str]


# magic function data
mfd_field = "mfd"


@dataclass
class MagicFunction:
    type: str
    expected_fun_args = None

    original_caller: Optional[Callable] = None

    def __new__(cls, *args, **kwargs) -> Callable:
        if args and callable(args[0]):
            fun = cast(Callable[..., Any], args[0])

            args = args[1:]

            cls._inject_data(fun, *args, **kwargs)
            return fun
        else:
            def decor(fun):
                cls._inject_data(fun, *args, **kwargs)
                return fun

            return decor

    @classmethod
    def _wrap_fun(cls, fun: Callable) -> Callable:
        return fun

    @classmethod
    def _inject_data(cls, wrapped: Callable, *args, **kwargs) -> None:
        wrapped.mfd = MagicFunctionData()
        wrapped.mfd.type = cls.type
        wrapped.mfd.expected_fun_args = cls.expected_fun_args

    @contextmanager
    def _context(self, *args, **kwargs) -> None:
        yield


@dataclass
class command(MagicFunction):
    type = "command"

    def __new__(cls, *args, in_root: Optional[bool] = True, cd_back: Optional[bool] = True) -> Callable:
        if args and callable(args[0]):
            fun = cast(Callable[..., Any], args[0])
            args = args[1:]

            @wraps(fun)
            def wrapped(*fun_args, **fun_kwargs):
                return cls._call(fun, fun_args, fun_kwargs, in_root, cd_back)

            cls._inject_data(wrapped, *args)
            return wrapped
        else:

            def decor(fun):
                @wraps(fun)
                def wrapped(*fun_args, **fun_kwargs):
                    return cls._call(fun, fun_args, fun_kwargs, in_root, cd_back)

                cls._inject_data(wrapped, in_root, cd_back)
                return wrapped

            return decor

    @classmethod
    def _call(cls, fun: Callable, fun_args, fun_kwargs, in_root: Optional[bool] = True, cd_back: Optional[bool] = True):
        if not cls.original_caller:
            cls.original_caller = fun

        if not fun_args or not isinstance(fun_args[0], Env):
            if fun_args and fun_args[0] == "__env__":
                fun_args = (builtins.__env__, *fun_args[1:])
            else:
                fun_args = (builtins.__env__, *fun_args)
        try:
            with cls._context(fun_args[0], in_root, cd_back):
                ret = fun(*fun_args, **fun_kwargs)
            return ret
        except Exception as e:
            if cls.original_caller is fun:
                logger.traceback()
                sys.exit(1)
            else:
                raise e
        finally:
            if cls.original_caller is fun:
                cls.original_caller = None

    @classmethod
    @contextmanager
    def _context(cls, env: "Env", in_root: Optional[bool] = True, cd_back: Optional[bool] = True) -> None:
        cwd = Path(".").absolute()

        if in_root:
            os.chdir(str(env.meta.root))

        try:
            yield
        finally:
            if cd_back:
                os.chdir(str(cwd))


class boot_code(MagicFunction):  # noqa: N801
    type: str = "boot_code"


class Event(MagicFunction):  # noqa: N801
    pass


class onload(Event):  # noqa: N801
    type: str = "onload"


class oncreate(Event):  # noqa: N801
    type: str = "oncreate"


class ondestroy(Event):  # noqa: N801
    type: str = "ondestroy"


class onunload(Event):  # noqa: N801
    type: str = "onunload"


class cmd_hook(MagicFunction):  # noqa: N801
    def __new__(cls, cmd_regex: str = ".*") -> Callable:
        ret = super().__new__(cls, cmd_regex)
        return ret

    @classmethod
    def _inject_data(cls, wrapped: Callable, cmd_regex: str = ".*") -> None:
        super()._inject_data(wrapped, cmd_regex)
        wrapped.mfd.cmd_regex = cmd_regex


class precmd(cmd_hook):  # noqa: N801
    type: str = "precmd"
    expected_fun_args = ["command", "out"]


class onstdout(cmd_hook):  # noqa: N801
    type: str = "onstdout"
    expected_fun_args = ["command", "out"]


class onstderr(cmd_hook):  # noqa: N801
    type: str = "onstderr"
    expected_fun_args = ["command", "out"]


class postcmd(cmd_hook):  # noqa: N801
    type: str = "postcmd"
    expected_fun_args = ["command", "stdout", "stderr"]


class shell_context(MagicFunction):  # noqa: N801
    type: str = "shell_context"

    def __init__(self) -> None:
        super().__init__()


PathLike = Union[Path, str]

magic_functions = {
    "command": command,
    "shell_context": shell_context,
    "boot_code": boot_code,
    "onload": onload,
    "onunload": onunload,
    "oncreate": oncreate,
    "ondestroy": ondestroy,
    "precmd": precmd,
    "onstdout": onstdout,
    "onstderr": onstderr,
}


@dataclass
class Source:
    root: Path
    watch_files: List[str] = field(default_factory=list)
    ignore_files: List[str] = field(default_factory=list)

    def __post_init__(self) -> None:
        self.root = self.root.resolve()


class BaseEnv(ABC):
    class Meta:
        pass

    @abstractmethod
    def init(self) -> None:
        pass

    @abstractmethod
    def post_init(self) -> None:
        pass


class Env(BaseEnv):
    class Meta:
        """
        Environment metadata.
        """

        root: Path
        name: Optional[str] = None
        version: str = "0.1.0"
        parents: List[str] = []
        plugins: List["Plugin"] = []
        sources: List[Source] = []
        emoji: str = ""
        stage: str = "comm"
        watch_files: List[str] = []
        ignore_files: List[str] = []
        verbose_run: bool = True
        load_env_vars: bool = False

    class Environ(envium.Environ):
        pythonpath: Optional[List[PathLike]] = env_var(raw=True, default_factory=list)
        path: Optional[List[PathLike]] = env_var(raw=True, default_factory=list)
        root: Optional[Path] = env_var()
        stage: Optional[str] = env_var()
        envo_stage: Optional[str] = env_var(raw=True)
        envo_name: Optional[str] = env_var(raw=True)

    class Ctx(envium.Ctx):
        pass

    class Secrets(envium.Secrets):
        pass

    ctx: Ctx
    secrets: Secrets
    meta: Meta

    _environ_before: Optional[Dict[str, str]]

    env_id_to_secrets: ClassVar[Dict[str, Secrets]] = {}
    _shell: "Shell"

    def __init__(self):
        self._environ_before = os.environ.copy()
        self.meta = self.Meta()

        self.e = self.Environ(name=self.meta.name, load=self.meta.load_env_vars)
        self.e.envo_name = self.meta.name

        self.e.root = self.meta.root
        self.e.stage = self.meta.stage
        self.e.envo_stage = self.meta.stage

        self.e.path = self._path_str_to_list(os.environ["PATH"])

        if "PYTHONPATH" not in os.environ:
            self.e.pythonpath = []
        else:
            self.e.pythonpath = self._path_str_to_list(os.environ["PYTHONPATH"])

        self.ctx = self.Ctx(self.meta.name)

        secrets = Env.env_id_to_secrets.get(self.id, self.Secrets(self.meta.name))
        self.secrets = Env.env_id_to_secrets[self.id] = secrets

        self.init()

        self.validate()
        self.activate()

        self.post_init()

    def _get_path_delimiter(self) -> str:
        if utils.is_linux() or utils.is_darwin():
            return ":"
        elif utils.is_windows():
            return ";"
        else:
            raise NotImplementedError

    def _path_str_to_list(self, path: str) -> List[Path]:
        paths_str = path.split(self._get_path_delimiter())
        ret = [Path(s) for s in paths_str]
        return ret

    @classmethod
    def instantiate(cls, stage: Optional[str] = None) -> "Env":
        if not stage:
            stage = environ.stage

        env_class = import_from_file(cls.Meta.root / f"env_{stage}.py").ThisEnv
        return env_class()

    def init(self) -> None:
        super().init()
        pass

    def post_init(self) -> None:
        pass

    @property
    def id(self) -> str:
        return f"{self.__class__.__module__}:{self.__class__.__name__}"

    def validate(self) -> None:
        """
        Validate env
        """

        msgs = []
        if self.ctx.errors:
            msgs.append(f"Context errors in {self.meta.root}:\n" + f"\n".join([str(e) for e in self.ctx.errors]))

        if self.e.errors:
            msgs.append(f"Environ errors in {self.meta.root}:\n" + f"\n".join([str(e) for e in self.e.errors]))

        if self.secrets.errors:
            msgs.append(f"Secrets errors  in {self.meta.root}:\n" + f"\n".join([str(e) for e in self.e.errors]))

        msg = "\n".join(msgs)

        if msg:
            raise UserError(msg)

    def get_env_vars(self) -> Dict[str, str]:
        ret = self.e.get_env_vars()
        return ret

    def get_user_envs(self) -> List[Type["Env"]]:
        ret = []
        for c in self.__class__.__mro__:
            if not issubclass(c, Env) or c is Env:
                continue
            ret.append(c)

        return ret

    def get_env(self, directory: Union[Path, str], stage: Optional[str] = None) -> "Env":
        stage = stage or self.meta.stage
        directory = Path(directory)
        env_file = directory / f"env_{stage}.py"

        if not env_file.exists():
            with user_code():
                logger.traceback()
            raise UserError(f"{env_file} does not exit")

        env = import_env_from_file(env_file).ThisEnv()
        return env

    @classmethod
    def get_env_path(cls) -> Path:
        return cls.Meta.root / f"env_{cls.Meta.stage}.py"

    def dump_dot_env(self) -> Path:
        """
        Dump .env file for the current environment.

        File name follows env_{env_name} format.
        """
        path = Path(f".env_{self.meta.stage}")
        content = "\n".join([f'{key}="{value}"' for key, value in self.e.get_env_vars().items()])
        path.write_text(content, "utf-8")
        return path

    def activate(self) -> None:
        if not self._environ_before:
            self._environ_before = os.environ.copy()

        os.environ.update(**self.get_env_vars())

    def deactivate(self) -> None:
        if self._environ_before:
            os.environ = self._environ_before.copy()


class ShellEnv:
    """
    Defines environment.
    """

    @dataclass
    class _Links:
        shell: Optional["Shell"]
        env: Env
        status: "Status"

    @dataclass
    class _Sets:
        blocking: bool = False

    _sys_modules_snapshot: Dict[str, ModuleType] = OrderedDict()
    magic_functions: Dict[str, Any]
    env: Env

    def __init__(self, se: _Sets, li: _Links) -> None:
        self._se = se
        self._li = li
        self.env = self._li.env

        self.env._shell = self._li.shell

        self.magic_functions = {
            "shell_context": {},
            "precmd": {},
            "onstdout": {},
            "onstderr": {},
            "postcmd": {},
            "onload": {},
            "oncreate": {},
            "ondestroy": {},
            "onunload": {},
            "boot_code": {},
            "command": {},
        }

        if self.env.meta.verbose_run:
            os.environ["ENVO_VERBOSE_RUN"] = "True"
        elif os.environ.get("ENVO_VERBOSE_RUN"):
            os.environ.pop("ENVO_VERBOSE_RUN")

        self._exiting = False

        self._shell_environ_before = None

        self.logger: Logger = logger.create_child("envo", descriptor=self.env.meta.name)

        self._environ_before = None
        self._shell_environ_before = None
        self._collect_magic_functions()

        self.logger.debug("Starting env", metadata={"root": self.env.meta.root, "stage": self.env.meta.stage})

        self._li.shell.calls.pre_cmd = Callback(self._on_precmd)
        self._li.shell.calls.on_stdout = Callback(self._on_stdout)
        self._li.shell.calls.on_stderr = Callback(self._on_stderr)
        self._li.shell.calls.post_cmd = Callback(self._on_postcmd)
        self._li.shell.calls.on_exit = Callback(self._on_destroy)

        if "" in sys.path:
            sys.path.remove("")

        if not self._sys_modules_snapshot:
            self._sys_modules_snapshot = OrderedDict(sys.modules.copy())

        builtins.__env__ = self.env

    def get_name(self) -> str:
        """
        Return env name
        """
        return self.env.meta.name

    def redraw_prompt(self) -> None:
        self._li.shell.redraw()

    def load(self) -> None:
        """
        Called after creation.
        :return:
        """

        def thread(self: "ShellEnv") -> None:
            logger.debug("Starting onload thread")

            sw = Stopwatch()
            sw.start()

            onload_funs = self.magic_functions["onload"].values()

            try:
                for h in onload_funs:
                    h(self.env)
                self._run_boot_codes()
            except Exception as e:
                with user_code():
                    logger.traceback()
                self._exit()
                return

            # declare commands
            for name, c in self.magic_functions["command"].items():
                self._li.shell.set_variable(name, c)

            # set context
            self._li.shell.set_context(self._get_shell_context())
            while sw.value <= 0.5:
                sleep(0.1)

            logger.debug("Finished load context thread")
            self._li.status.shell_context_ready = True
            self.first_run = False

        if not self._se.blocking:
            Thread(target=thread, args=(self,)).start()
        else:
            thread(self)

    def on_shell_create(self) -> None:
        """
        Called only after creation.
        :return:
        """
        functions = self.magic_functions["oncreate"].values()
        for h in functions:
            h(self.env)

    def _get_namespace_and_function(self, name: str) -> Tuple[Optional[str], str]:
        if "__" in name and not name.endswith("__"):
            namespace = name.split("__")[0]
            f = "".join(name.split("__")[1:])
            return namespace, f
        else:
            return None, name

    def _collect_magic_functions(self) -> None:
        """
        Go through fields and transform decorated functions to commands.
        """

        def hasattr_static(obj: Any, field: str) -> bool:
            try:
                inspect.getattr_static(obj, field)
            except AttributeError:
                return False
            else:
                return True

        for c in reversed(self.env.__class__.__mro__):
            for f in dir(c):
                if hasattr_static(self.__class__, f) and inspect.isdatadescriptor(
                    inspect.getattr_static(self.__class__, f)
                ):
                    continue

                attr = inspect.getattr_static(c, f)

                if hasattr(attr, mfd_field):
                    namespace, f = self._get_namespace_and_function(f)
                    namespaced_name = f"{namespace}.{f}" if namespace else f
                    self.magic_functions[attr.mfd.type][namespaced_name] = attr

    def _get_shell_context(self) -> Dict[str, Any]:
        shell_context = {}
        for c in self.magic_functions["shell_context"].values():
            try:
                with user_code():
                    cont = c(self.env)
            except Exception as e:
                logger.traceback()
                continue
            for k, v in cont.items():
                namespace, f = self._get_namespace_and_function(k)

                namespaced_name = f"{namespace}.{f}" if namespace else k
                shell_context[namespaced_name] = v

        return shell_context

    def _on_destroy(self) -> None:
        functions = self.magic_functions["ondestroy"]
        for h in functions.values():
            h()

        self._exit()

    def _exit(self) -> None:
        self.logger.debug("Exiting env")

    def activate(self) -> None:
        """
        Validate env and send vars to os.environ

        :param owner_namespace:
        """
        if not self._shell_environ_before:
            self._shell_environ_before = dict(self._li.shell.environ.items())
        self._li.shell.environ.update(**self.env.get_env_vars())

    def _deactivate(self) -> None:
        """
        Validate env and send vars to os.environ

        :param owner_namespace:
        """
        if self._shell_environ_before:
            if self._li.shell:
                tmp_environ = copy(self._li.shell.environ)
                for i, v in tmp_environ.items():
                    self._li.shell.environ.pop(i)
                for k, v in self._shell_environ_before.items():
                    if v is None:
                        continue
                    self._li.shell.environ[k] = v

        self.env.deactivate()

    def _run_boot_codes(self) -> None:
        self._li.status.source_ready = False
        boot_codes_f = self.magic_functions["boot_code"]

        codes = []

        for f in boot_codes_f.values():
            with user_code():
                codes.extend(f(self.env))

        for c in codes:
            self._li.shell.run_code(c)

        self._li.status.source_ready = True

    def _on_precmd(self, command: str) -> Optional[str]:
        functions = self.magic_functions["precmd"]
        for f in functions.values():
            if re.match(f.mfd.cmd_regex, command):
                with user_code():
                    ret = f(self.env, command)  # type: ignore
                command = ret

        return command

    def _on_stdout(self, command: str, out: bytes) -> str:
        functions = self.magic_functions["onstdout"]
        for f in functions.values():
            if re.match(f.mfd.cmd_regex, command):
                with user_code():
                    ret = f(self.env, command, out)  # type: ignore
                if ret:
                    out = ret
        return out

    def _on_stderr(self, command: str, out: bytes) -> str:
        functions = self.magic_functions["onstderr"]
        for f in functions.values():
            if re.match(f.mfd.cmd_regex, command):
                with user_code():
                    ret = f(self.env, command, out)  # type: ignore
                if ret:
                    out = ret
        return out

    def _on_postcmd(self, command: str, stdout: str, stderr: str) -> None:
        functions = self.magic_functions["postcmd"]
        for f in functions.values():
            if re.match(f.mfd.cmd_regex, command):
                with user_code():
                    f(self.env, command, stdout, stderr)  # type: ignore

    def _unload(self) -> None:
        self._deactivate()
        functions = self.magic_functions["onunload"]
        for f in functions.values():
            f(self.env)
        self._li.shell.calls.reset()
